#include "PhysicsHandler.h"
#include "NxMeshCooker.h"

using namespace Ogre;

PhysicsHandler::PhysicsHandler(Ogre::SceneManager* msceneManager, GameController* _game){
	sceneManager = msceneManager;
	LogManager::getSingleton().logMessage("Initializing PhysX engine...");
	inited = false;
	physicsSDK = NxCreatePhysicsSDK(NX_PHYSICS_SDK_VERSION);
	if (!physicsSDK)  return;

	physicsSDK->getFoundationSDK().getRemoteDebugger()->connect("localhost", 5425);

	physicsSDK->setParameter(NX_SKIN_WIDTH, 0.1);
	NxSceneDesc sceneDesc;
	sceneDesc.simType = NX_SIMULATION_SW;
	sceneDesc.gravity = NxVec3(0,-50,50);
	sceneDesc.groundPlane = false; //true;

	physicsScene = physicsSDK->createScene(sceneDesc);

	if(!physicsScene){ 
		LogManager::getSingleton().logMessage("Hardware physX simulation not supported, switching to software");
		sceneDesc.simType = NX_SIMULATION_SW; 
		physicsScene = physicsSDK->createScene(sceneDesc);  
		if(!physicsScene) return;
	}
	
	// Create the default material
	NxMaterial* defaultMaterial = physicsScene->getMaterialFromIndex(0); 
	defaultMaterial->setRestitution(0.5);
	defaultMaterial->setStaticFriction(0.5);
	defaultMaterial->setDynamicFriction(0.5);

	tablebase = new TableBase(sceneManager, physicsScene, physicsSDK);
	lever = new Lever(sceneManager, physicsScene, physicsSDK);
	triggerable = new Triggerable(sceneManager, physicsScene, physicsSDK, _game);
	balls.push_back(new Ball(sceneManager, physicsScene, physicsSDK, 0));
	inited = true;
}

PhysicsHandler::~PhysicsHandler(){}

bool PhysicsHandler::frameStarted(const FrameEvent& evt){
	if(!inited)
		return true;

	physicsScene->simulate(evt.timeSinceLastFrame);
	physicsScene->flushStream();
	while (!physicsScene->fetchResults(NX_RIGID_BODY_FINISHED, false));

	std::vector<Ball*>::iterator it = balls.begin();
	std::vector<Ball*>::iterator itend = balls.end();
	while(it != itend)
	{
		(*it)->update();
		it++;
	}
	lever->update();

	return true;
}
void PhysicsHandler::resetBall(){
	balls[0]->reset();
}
void PhysicsHandler::startBall(){
	balls[0]->start();
}
void PhysicsHandler::travelBall(int time){
	balls[0]->travel(time);
}
Lever* PhysicsHandler::getLever(){
	return lever;
}

Triggerable::Triggerable(Ogre::SceneManager* sceneManager, NxScene* mPhysicsScene, NxPhysicsSDK* mPhysicsSDK, GameController* _game){
	game = _game;

	// booster
	booster = sceneManager->getRootSceneNode()->createChildSceneNode();
	Entity* boosterEnt = sceneManager->createEntity("booster", "booster.mesh");
	//booster->attachObject(boosterEnt);
	//booster->setScale(1,1,1);
	//booster->setPosition(0,0,0);
	NxMeshCooker::cookNxMesh(boosterEnt->getMesh(), UserStream("media/booster.mesh.nxs", false), Vector3(1,1,1));
	boosterShapeDesc.meshData = mPhysicsSDK->createTriangleMesh(UserStream("media/booster.mesh.nxs", true));
	boosterShapeDesc.shapeFlags |= NX_TRIGGER_ENABLE;
	boosterActorDesc.shapes.pushBack(&boosterShapeDesc);
	boosterActorDesc.globalPose.t = NxVec3(0,0,0);
	boosterNx = mPhysicsScene->createActor(boosterActorDesc);
	boosterNx->setName("booster");

	// hole
	hole = sceneManager->getRootSceneNode()->createChildSceneNode();
	Entity* holeEnt = sceneManager->createEntity("hole", "hole.mesh");
	//hole->attachObject(holeEnt);
	//hole->setScale(1,1,1);
	//hole->setPosition(0,0,0);
	NxMeshCooker::cookNxMesh(holeEnt->getMesh(), UserStream("media/hole.mesh.nxs", false), Vector3(1,1,1));
	holeShapeDesc.meshData = mPhysicsSDK->createTriangleMesh(UserStream("media/hole.mesh.nxs", true));
	holeShapeDesc.shapeFlags |= NX_TRIGGER_ENABLE;
	holeActorDesc.shapes.pushBack(&holeShapeDesc);
	holeActorDesc.globalPose.t = NxVec3(0,0,0);
	holeNx = mPhysicsScene->createActor(holeActorDesc);
	holeNx->setName("hole");

	// springs
	springl = sceneManager->getRootSceneNode()->createChildSceneNode();
	Entity* springlEnt = sceneManager->createEntity("springleft", "springleft.mesh");
	//springl->attachObject(springlEnt);
	//springl->setScale(1,1,1);
	//springl->setPosition(0,0,0);
	NxMeshCooker::cookNxMesh(springlEnt->getMesh(), UserStream("media/springleft.mesh.nxs", false), Vector3(1,1,1));
	springlShapeDesc.meshData = mPhysicsSDK->createTriangleMesh(UserStream("media/springleft.mesh.nxs", true));
	springlShapeDesc.shapeFlags |= NX_TRIGGER_ENABLE;
	springlActorDesc.shapes.pushBack(&springlShapeDesc);
	springlActorDesc.globalPose.t = NxVec3(0,0,0);
	springlNx = mPhysicsScene->createActor(springlActorDesc);
	springlNx->setName("springleft");

	springr = sceneManager->getRootSceneNode()->createChildSceneNode();
	Entity* springrEnt = sceneManager->createEntity("springright", "springright.mesh");
	//springr->attachObject(springrEnt);
	//springr->setScale(1,1,1);
	//springr->setPosition(0,0,0);
	NxMeshCooker::cookNxMesh(springrEnt->getMesh(), UserStream("media/springright.mesh.nxs", false), Vector3(1,1,1));
	springrShapeDesc.meshData = mPhysicsSDK->createTriangleMesh(UserStream("media/springright.mesh.nxs", true));
	springrShapeDesc.shapeFlags |= NX_TRIGGER_ENABLE;
	springrActorDesc.shapes.pushBack(&springrShapeDesc);
	springrActorDesc.globalPose.t = NxVec3(0,0,0);
	springrNx = mPhysicsScene->createActor(springrActorDesc);
	springrNx->setName("springright");

	// buttons	
	buttons = sceneManager->getRootSceneNode()->createChildSceneNode();
	Entity* buttonsEnt = sceneManager->createEntity("buttons", "buttons.mesh");
	buttons->attachObject(buttonsEnt);
	buttons->setScale(1,1,1);
	buttons->setPosition(0,0,0);

	//NxMeshCooker::cookNxMesh(buttonsEnt->getMesh(), UserStream("media/buttons.mesh.nxs", false), Vector3(1,1,1));
	//buttonsShapeDesc.meshData = mPhysicsSDK->createTriangleMesh(UserStream("media/buttons.mesh.nxs", true));
	for(int i=0; i<8; i++){
		NxSphereShapeDesc tmpdes;
		tmpdes.radius = 2.5;
		switch (i){
			case 0:
				tmpdes.localPose.t.set(NxVec3(-17,-1,-28.5));
				break;
			case 1:
				tmpdes.localPose.t.set(NxVec3(-1.7,-1,-28.5));
				break;
			case 2:
				tmpdes.localPose.t.set(NxVec3(14,-1,-28.5));
				break;
			case 3:
				tmpdes.localPose.t.set(NxVec3(6.7,-1,-22.2));
				break;
			case 4:
				tmpdes.localPose.t.set(NxVec3(-9,-1,-22.2));
				break;
			case 5:
				tmpdes.localPose.t.set(NxVec3(5.7,-1,9));
				break;
			case 6:
				tmpdes.localPose.t.set(NxVec3(16.6,-1,9));
				break;
			case 7:
				tmpdes.localPose.t.set(NxVec3(11.1,-0,24.8));
				break;
		}
		tmpdes.shapeFlags |= NX_TRIGGER_ENABLE;
		buttonsActorDesc.shapes.pushBack(&tmpdes);
		buttonsActorDesc.globalPose.t.set(NxVec3(0,0,0));
		buttonsNx = mPhysicsScene->createActor(buttonsActorDesc);
		buttonsNx->setName("buttons");
	}


	// time travel
	time1955 = sceneManager->getRootSceneNode()->createChildSceneNode();
	Entity* time1955Ent = sceneManager->createEntity("time1955", "timeleft.mesh");
	//time1955->attachObject(time1955Ent);
	//time1955->setScale(1,1,1);
	//time1955->setPosition(0,0,0);
	NxMeshCooker::cookNxMesh(time1955Ent->getMesh(), UserStream("media/timeleft.mesh.nxs", false), Vector3(1,1,1));
	time1955ShapeDesc.meshData = mPhysicsSDK->createTriangleMesh(UserStream("media/timeleft.mesh.nxs", true));
	time1955ShapeDesc.shapeFlags |= NX_TRIGGER_ENABLE;
	time1955ActorDesc.shapes.pushBack(&time1955ShapeDesc);
	time1955ActorDesc.globalPose.t = NxVec3(0,0,0);
	time1955Nx = mPhysicsScene->createActor(time1955ActorDesc);
	time1955Nx->setName("timeleft");

	time1985 = sceneManager->getRootSceneNode()->createChildSceneNode();
	Entity* time1985Ent = sceneManager->createEntity("timeright", "timeright.mesh");
	//time1985->attachObject(time1985Ent);
	//time1985->setScale(1,1,1);
	//time1985->setPosition(0,0,0);
	NxMeshCooker::cookNxMesh(time1985Ent->getMesh(), UserStream("media/timeright.mesh.nxs", false), Vector3(1,1,1));
	time1985ShapeDesc.meshData = mPhysicsSDK->createTriangleMesh(UserStream("media/timeright.mesh.nxs", true));
	time1985ShapeDesc.shapeFlags |= NX_TRIGGER_ENABLE;
	time1985ActorDesc.shapes.pushBack(&time1985ShapeDesc);
	time1985ActorDesc.globalPose.t = NxVec3(0,0,0);
	time1985Nx = mPhysicsScene->createActor(time1985ActorDesc);
	time1985Nx->setName("timeright");

	mPhysicsScene->setUserTriggerReport(this);
}

void Triggerable::onTrigger(NxShape& triggerShape, NxShape& otherShape, NxTriggerFlag status){
	NxActor* tmpact = &triggerShape.getActor();
	NxActor* ball = &otherShape.getActor();
	std::string name = tmpact->getName();
	NxVec3 velocity = ball->getLinearVelocity();
	if(velocity.dot(velocity) == 0)
		return;

	float powerx = 0;
	float powery = 0;
	float powerz = 0;

	if(name == "booster"){
		powerx = 0;
		powery = 0;
		powerz = (-3000) + (-(rand()%2000));
		ball->addForce(NxVec3(powerx,powery,powerz), NX_ACCELERATION);		
	}
	else if(name == "springleft"){
		game->addPoints(50);
		powerx = 300 + (rand()%500);
		powery = 0;
		powerz = (-500) + (-(rand()%1000));
		ball->addForce(NxVec3(powerx,powery,powerz), NX_ACCELERATION);
	}
	else if(name == "springright"){
		game->addPoints(50);
		powerx = (-300) + (-(rand()%500));
		powery = 0;
		powerz = (-500) + (-(rand()%1000));
		ball->addForce(NxVec3(powerx,powery,powerz), NX_ACCELERATION);
	}
	else if(name == "hole" && (status & NX_TRIGGER_ON_ENTER)){
		game->setGamestate("* The ball disintegrated *", 2.0);
		game->resetBall();
	}
	if(name == "timeleft" && (status & NX_TRIGGER_ON_ENTER)){
		game->timeTravel(1955);		
	}
	if(name == "timeright" && (status & NX_TRIGGER_ON_ENTER)){
		game->timeTravel(1985);		
	}
	else if(name == "buttons" && (status & NX_TRIGGER_ON_ENTER)){
		game->addPoints(100);
		LogManager::getSingleton().logMessage("x:"+StringConverter::toString(velocity.x)+" y:"+StringConverter::toString(velocity.y)+" z:"+StringConverter::toString(velocity.z));
		velocity.normalize();
		LogManager::getSingleton().logMessage("x:"+StringConverter::toString(velocity.x)+" y:"+StringConverter::toString(velocity.y)+" z:"+StringConverter::toString(velocity.z));

		/*
		int power = 500;
		int randomx = (rand()%200);
		int randomy = (rand()%200);
		powerx = velocity.x*10*power*(-1)+randomx;
		powery = 0;
		powerz = velocity.z*10*power*(-1)+randomy;
		*/

		float power = 500;
		float randomx = (rand()%200);
		float randomz = (rand()%200);
		powerx = velocity.x*(-1)*power;
		powery = 0;
		powerz = velocity.z*(-1)*power;

		if(100*abs(velocity.x)<abs(velocity.z))
			if(powerx>0)
				powerx = powerx + randomx;
			else
				powerx = powerx - randomx;

		if(abs(powerx)<200)
			if(powerx>0)
				powerx = powerx + (rand()%50)+50;
			else
				powerx = powerx - (rand()%50)+50;
		if(abs(powerz)<200)
			if(powerz>0)
				powerz = powerz + (rand()%50)+50;
			else
				powerz = powerz - (rand()%50)+50;

		ball->setLinearVelocity(NxVec3(0,0,0));
		NxVec3 trigPos = triggerShape.getLocalPosition();
		NxVec3 ballPos = ball->getGlobalPosition();
		NxVec3 modVec = ballPos - trigPos;
		modVec.normalize();
		modVec.y = 0;
		ball->setGlobalPosition(trigPos+modVec*3.6);
		LogManager::getSingleton().logMessage("x:"+StringConverter::toString(powerx)+" y:"+StringConverter::toString(powery)+" z:"+StringConverter::toString(powerz));
		ball->addForce(modVec*5000, NX_ACCELERATION);
	}

}
